;;;-*- Mode: Lisp; Package: OCI -*-

;; OCI module
;; Version 0.5 (experimental)
;; Copyright (C) Paul Meurer 2000. All rights reserved.
;; paul.meurer@hit.uib.no
;;
;; Documentation and the license agreement can be found in file 
;; "sql-odbc-documentation.lisp".
;; Bug reports and suggestions are highly welcome.

(in-package :oci)
;;(use-package :ffc)

#+lispworks
(progn
  #-unix
  (fli:register-module "oci" :connection-style :automatic)
  #+unix
  (fli:register-module "/opt/oracle/OraHome1/lib/libclntsh.so"
                       :connection-style :automatic)

  (fli:define-c-typedef ub2 :unsigned-short)
  (fli:define-c-typedef sb2 :signed-short) ; :short
  (fli:define-c-typedef ub4 :unsigned-long)
  (fli:define-c-typedef sb4 :signed-long) ; :long
  (fli:define-c-typedef uword :unsigned-long)
  (fli:define-c-typedef dvoid (:pointer :void))
  (fli:define-c-typedef size-t :unsigned-integer) ; :unsigned-int
  ;; we make life easy and reuse definitions made for odbc ;-)
  ;;(fli:define-c-typedef sql-handle (:pointer :void))
  ;;(fli:define-c-typedef sql-handle-ptr (:pointer sql-handle))

  (fli:define-c-typedef OCIEnv (:pointer :void)) ; ??
  (fli:define-c-typedef OCIErr (:pointer :void)) ; ??
  ;;(fli:define-c-typedef text (:pointer (:unsigned :char)))
  (fli:define-c-typedef text string-ptr))

#+allegro
(eval-when (:execute :load-toplevel :compile-toplevel)
  #-unix
  (setf *foreign-module* "oci.dll")
  #+unix
  (setf *foreign-module* "/opt/oracle/OraHome1/lib/libclntsh.so"))


(define-foreign-function "OCIInitialize"
    ((mode :unsigned-long)
     (ctxp :pointer)
     (malocfp :pointer)
     (ralocfp :pointer)
     (mfreefp :pointer)
     )
  :signed-short)

(define-foreign-function "OCIHandleAlloc"
    ((parenth :pointer)
     (hndlpp :pointer)
     (type :unsigned-long)
     (xtramem-sz :unsigned-int)
     (usrmempp :pointer)
     )
  :signed-short)

(define-foreign-function "OCIHandleFree"
    ((hndlpp :pointer)
     (type :unsigned-long)
     )
  :signed-short)

(define-foreign-function "OCIEnvInit"
    ((envp :pointer)
     (mode :unsigned-long)
     (xtramem-sz :unsigned-int)
     (usrmempp :pointer)
     )
  :signed-short)

(define-foreign-function "OCIServerAttach"
    ((srvhp :pointer)
     (errhp :pointer)
     (dblink string-ptr)
     (dblink-len :signed-long)
     (mode :unsigned-long)
     )
  :signed-short)

(define-foreign-function "OCIServerDetach"
    ((srvhp :pointer)
     (errhp :pointer)
     (mode :unsigned-long)
     )
  :signed-short)

(define-foreign-function "OCISessionBegin"
    ((srvhp :pointer)
     (errhp :pointer)
     (usrhp :pointer)
     (credt :unsigned-long)
     (mode :unsigned-long)
     )
  :signed-short)

(define-foreign-function "OCISessionEnd"
    ((svchp :pointer)
     (errhp :pointer)
     (usrhp :pointer)
     (mode :unsigned-long)
     )
  :signed-short)
(define-foreign-function "OCIStmtPrepare"
    ((stmtp :pointer)
     (errhp :pointer)
     (stmt string-ptr)
     (stmt-len :unsigned-long)
     (language :unsigned-long)
     (mode :unsigned-long)
     )
  :signed-short)

(define-foreign-function "OCIStmtExecute"
    ((svchp :pointer)
     (stmtp :pointer)
     (errhp :pointer)
     (iters :unsigned-long)
     (rowoff :unsigned-long)
     (snap-in :pointer)
     (snap-out :pointer)
     (mode :unsigned-long)
     )
  :signed-short)

(define-foreign-function "OCIDefineByPos"
    ((stmtp :pointer)
     (defnp :pointer)
     (errhp :pointer)
     (position :unsigned-long)
     (valuep :pointer)
     (value-sz :signed-long)
     (dty :unsigned-short)
     (indp :pointer)
     (rlenp :pointer #+ignore(:pointer :unsigned-short))
     (rcodep :pointer #+ignore(:pointer :unsigned-short))
     (mode :unsigned-long)
     )
  :signed-short)

(define-foreign-function "OCIDefineArrayOfStruct"
    ((defnp :pointer)
     (errhp :pointer)
     (pvskip :unsigned-long)
     (indskip :unsigned-long)
     (rlskip :unsigned-long)
     (rcskip :unsigned-long)
     )
  :signed-short)

(define-foreign-function "OCIStmtFetch"
    ((stmtp :pointer)
     (errhp :pointer)
     (nrows :unsigned-short) ;; **
     (orientation :unsigned-short)
     (mode :unsigned-short) ;; **
     )
  :signed-short)

(define-foreign-function "OCIParamGet"
    ((hndlp :pointer)
     (htype :unsigned-long)
     (errhp :pointer)
     (parmdpp :pointer)
     (pos :unsigned-long)
     )
  :signed-short)

(define-foreign-function "OCITransStart"
    ((svchp :pointer)
     (errhp :pointer)
     (timeout :unsigned-long #+ignore uword)
     (flags :unsigned-long)
     )
  :signed-short)

(define-foreign-function "OCITransCommit"
    ((svchp :pointer)
     (errhp :pointer)
     (flags :unsigned-long)
     )
  :signed-short)

(define-foreign-function "OCITransRollback"
    ((svchp :pointer)
     (errhp :pointer)
     (flags :unsigned-long)
     )
  :signed-short)

(define-foreign-function "OCIErrorGet"
    ((hndlp :pointer)
     (recordno :unsigned-long)
     (sqlstate :pointer #+ignore string-ptr)
     (errcodep :pointer)
     (bufp #+allegro :ptr
	   #+lispworks string-ptr)
     (bufsiz :unsigned-long)
     (type :unsigned-long)
     )
  :signed-short)

(define-foreign-function "OCIAttrGet"
    ((trgthndlp :pointer)
     (trghndltyp :unsigned-long)
     (attributep :pointer)
     (size :pointer)
     (attrtype :unsigned-long)
     (errhp :pointer)
     )
  :signed-short)

#+allegro
(define-foreign-function "OCIAttrGet"
    ((trgthndlp :pointer)
     (trghndltyp :unsigned-long)
     (attributep :ptr #+ignore string-ptr)
     (size :pointer)
     (attrtype :unsigned-long)
     (errhp :pointer)
     )
  :signed-short
  :lisp-name OCIAttrGet-string)

(define-foreign-function "OCIAttrSet"
    ((trgthndlp :pointer)
     (trghndltyp :unsigned-long)
     (attributep :pointer)
     (size :unsigned-long)
     (attrtype :unsigned-long)
     (errhp :pointer)
     )
  :signed-short)

#+allegro
(define-foreign-function "OCIAttrSet"
    ((trgthndlp :pointer)
     (trghndltyp :unsigned-long)
     (attributep string-ptr)
     (size :unsigned-long)
     (attrtype :unsigned-long)
     (errhp :pointer)
     )
  :signed-short
  :lisp-name OCIAttrSet-string)

#+oracle-i
(define-foreign-function "OCIEnvCreate"
    ((envhpp :pointer)
     (mode :unsigned-long)
     (ctxp :pointer)
     (malocfp :pointer)
     (ralocfp :pointer)
     (mfreefp :pointer)
     (xtramem-sz :unsigned-int)
     (usrmempp :pointer)
     )
  :signed-short)

:: EOF