;;;-*- Mode: Lisp; Package: OCI -*-

;; OCI module
;; Version 0.5 (experimemtal)
;; Copyright (C) Paul Meurer 2001. All rights reserved.
;; paul.meurer@hit.uib.no
;;

;; Documentation and the license agreement can be found in file 
;; "sql-odbc-documentation.lisp".
;; Bug reports and suggestions are highly welcome.

(in-package :oci)

;; should it rather be sleep??
(defparameter *wait-function*
  #+lispworks
  #'mp::process-allow-scheduling
  #+allegro
  #'mp::process-allow-schedule)

(defclass oci-query (query)
  ((oci-stmt :initform nil :initarg :oci-stmt :accessor oci-stmt)
   (statement-type :initform nil :accessor statement-type)
   (row-fetch-count :initform 1
		    :initarg :row-fetch-count :accessor row-fetch-count)
   (%no-data-p :initform nil :accessor %no-data-p)
   (prefetched-rows :initform (make-array 0 :adjustable t :fill-pointer t)
                    :accessor prefetched-rows)
   (column-count :initform nil :accessor column-count)
   (column-names :initform (make-array 0 :element-type 'string :adjustable t :fill-pointer t)
                 :accessor column-names)
   (column-sql-types :initform (make-array 0 :element-type 'fixnum :adjustable t :fill-pointer t)
                     :accessor column-sql-types)
   (column-def-ptrs :initform (make-array 0 :adjustable t :fill-pointer t)
                    :accessor def-ptrs)
   (column-data-ptrs :initform (make-array 0 :adjustable t :fill-pointer t)
                     :accessor data-ptrs)
   (column-out-len-ptrs :initform (make-array 0 :adjustable t :fill-pointer t)
                        :accessor column-out-len-ptrs)
   (column-return-code-ptrs :initform (make-array 0 :adjustable t :fill-pointer t)
                        :accessor column-return-code-ptrs)
   (column-precisions :initform (make-array 0 :element-type 'fixnum :adjustable t :fill-pointer t)
                      :accessor column-precisions)
   (column-data-sizes :initform (make-array 0 :element-type 'fixnum :adjustable t :fill-pointer t)
                      :accessor column-data-sizes)
   (column-scales :initform (make-array 0 :element-type 'fixnum :adjustable t :fill-pointer t)
                  :accessor column-scales)
   (column-nullables-p :initform (make-array 0 :element-type 'fixnum :adjustable t :fill-pointer t)
                       :accessor column-nullables-p)
   ;;(parameter-count :initform 0 :accessor parameter-count)
   (parameter-data-ptrs :initform (make-array 0 :adjustable t :fill-pointer t)
                     :accessor parameter-ptrs)))

(defclass oci-database (database sql::oracle-mixin)
  (;; any reason to have more than one oci-env?
   (oci-environment :initform nil ; :allocation :class
            :initarg :oci-environment :accessor oci-environment)
   (oci-error :initform nil :initarg :oci-error :accessor oci-error)
   (oci-service-context :initform nil :initarg :oci-service-context :accessor oci-service-context)
   (oci-server :initform nil :initarg :oci-server :accessor oci-server)
   (oci-session :initform nil :initarg :oci-session :accessor oci-session)
   (oci-transaction :initform nil :initarg :oci-transaction :accessor oci-transaction)
   ;; info ??
   (info :initform (make-hash-table) :reader db-info))
  #+cormanlisp (:metaclass cl::class-slot-class))

;; use this method if the db object does not exist yet

;; truely global (with class allocation):
;;  oci-environment ; is not necessarily global!
;; one for each connection:
;;  oci-error
;;  oci-service-context ; operational context for OCI calls
;;   oci-server  ; physical connection
;;   oci-session ; users' roles and privileges, operational context for call execution
;;   oci-transaction

(defmethod db-connect ((db-type (eql :oci)) db-name user-id password autocommit)
  (if nil ;*odbc-driver-connect-p* 
      (driver-connect db-type :db-name db-name :user-id user-id :password password)
    (let ((db (make-instance 'oci-database :db-type :oci
                             :db-name db-name
                             :user-id user-id
                             :password password)))
      (with-slots (oci-environment oci-error oci-service-context oci-server oci-session) db
        (unless oci-environment ;; has class allocation!
          (setf oci-environment (%new-environment-handle (%null-ptr))))
        (setf oci-error (%new-error-handle oci-environment)
              oci-service-context (%new-service-context-handle oci-environment oci-error)
              oci-server (%new-server-handle oci-environment oci-error))
        ;; if connection cannot be established, we drop out here.
        (db-connect db db-name user-id password autocommit)))))

(defmethod db-connect ((db oci-database) db-name user-id password autocommit)
  (with-slots (oci-environment oci-service-context oci-server oci-session oci-error) db
    (setf oci-session (%oci-connect oci-environment oci-server oci-service-context oci-error
                                    db-name user-id password)))
  #+ignore-yet
  (if autocommit
      (enable-autocommit (hdbc db))
    (disable-autocommit (hdbc db)))
  db)

(defmethod db-disconnect ((database oci-database))
  (with-slots (oci-service-context oci-server oci-error queries sql::connected-p) database
    (when sql::connected-p
      (dolist (query queries)
        (if t ;(query-active-p query)
            (with-slots (oci-stmt) query
              (when oci-stmt 
                (%free-statement oci-stmt oci-error)
                (setf oci-stmt nil)))))
      (%oci-disconnect oci-server oci-service-context oci-error)
      (let ((reset-default-db-p (eq *default-database* database)))
        (setf *connected-databases* (delete database *connected-databases*)
              sql::connected-p nil)
        (when reset-default-db-p
          (setf *default-database* (car *connected-databases*)))))))

(defmethod db-commit ((database oci-database))
  (with-slots (oci-service-context oci-error) database
    (%oci-trans-commit oci-service-context oci-error)))

(defmethod db-rollback ((database oci-database))
  (with-slots (oci-service-context oci-error) database
    (%oci-trans-rollback oci-service-context oci-error)))

(defmethod initialize-instance :after ((query oci-query)
                                       &key sql oci-env oci-error oci-svcctx &allow-other-keys)
  (when sql
    (let ((oci-stmt (%new-statement-handle oci-env oci-error)))
      ;;(%sql-exec-direct sql hstmt henv hdbc)
      (%oci-stmt-prepare oci-stmt oci-error sql)
      (%oci-stmt-execute oci-svcctx oci-stmt oci-error 1 *wait-function*)
      (setf (oci-stmt query) oci-stmt)
      (%initialize-query query)
      (with-slots (active-p) query
        (setf active-p t)))))

(defmethod db-fetch-query-results ((database oci-database) &optional count flatp)
  (db-fetch-query-results (sql::db-query-object database) count flatp))

(defmethod db-fetch-query-results ((query oci-query) &optional count flatp)
  (when (query-active-p query)
    (with-slots (column-count
                 prefetched-rows
                 row-fetch-count
                 column-data-ptrs column-sql-types 
                 column-out-len-ptrs column-return-code-ptrs column-precisions
                 column-data-sizes oci-stmt database)
        query
      (with-slots (oci-error) database
        (values
         (let ((result 0))
           (cond (flatp
                  (when (> column-count 1)
                    (error "If more than one column is to be fetched, flatp has to be nil."))
                  (let ((data-ptr (aref column-data-ptrs 0))
                        (sql-type (aref column-sql-types 0))
                        (out-len-ptr (aref column-out-len-ptrs 0))
                        (return-code-ptr (aref column-return-code-ptrs 0))
                        (precision (aref column-precisions 0))
                        (data-size (aref column-data-sizes 0)))
		    (if (= row-fetch-count 1)
		      (loop for i from 0
			    until (and count (= i count))
			    do (setf result
				     (%oci-stmt-fetch oci-stmt oci-error
						      row-fetch-count
                                                      *wait-function*))
			    collect
			    (cond ((< 0 data-size +max-precision+)
				   (read-data data-ptr sql-type out-len-ptr
					      return-code-ptr
					      data-size row-fetch-count))
				  ((zerop (%get-long out-len-ptr)) ; change to ub2
				   nil)
				  (t #+ignore-yet
				     (read-data-in-chunks hstmt 0 data-ptr c-type sql-type
							  out-len-ptr nil)))
			    until (and (or ; (> row-fetch-count 1)
					   (null (aref prefetched-rows 0)))
				       (= result +OCI-NO-DATA+)))
		      (loop for i from 0
			    until (and count (= i count))
			    do (when (null (aref prefetched-rows 0))
				 ;; reset return-code-ptr
				 (loop for return-code-ptr
				       across column-return-code-ptrs
				       do
				       (dotimes (i row-fetch-count)
					 (setf
					  #+allegro
					  (ff:fslot-value return-code-ptr i)
					  #+lispworks
					  (fli::foreign-aref return-code-ptr i)
					  +OCI-NO-DATA+)))
				 (setf result
				       (%oci-stmt-fetch oci-stmt oci-error
							row-fetch-count *wait-function*)))
			    append
			    (cond ((< 0 data-size +max-precision+)
				   (read-data data-ptr sql-type out-len-ptr return-code-ptr
					      data-size row-fetch-count))
				  ((zerop (%get-long out-len-ptr)) ; change to ub2
				   nil)
				  (t #+ignore-yet
				     (read-data-in-chunks hstmt 0 data-ptr c-type sql-type
							  out-len-ptr nil)))
			    until (and (or (> row-fetch-count 1)
					   (null (aref prefetched-rows 0)))
				       (= result +OCI-NO-DATA+))))))
                 ((= row-fetch-count 1)
                  (loop for i from 0
                        until (or (and count (= i count))
                                  (= (%oci-stmt-fetch oci-stmt oci-error row-fetch-count 
                                                      *wait-function*)
                                     +OCI-NO-DATA+))
                        collect
                        (loop for data-ptr across column-data-ptrs
                              for sql-type across column-sql-types
                              for out-len-ptr across column-out-len-ptrs
                              for return-code-ptr across column-return-code-ptrs
                              for precision across column-precisions
                              for data-size across column-data-sizes
                              for j from 0 ; column count is zero based in lisp
                              collect 
                              (cond ((< 0 data-size +max-precision+)
                                     (read-data data-ptr sql-type
                                                out-len-ptr return-code-ptr
                                                data-size row-fetch-count))
                                    ((zerop (%get-long out-len-ptr))
                                     nil)
                                    (t #+ignore-yet
                                       (read-data-in-chunks hstmt j data-ptr c-type sql-type
                                                            out-len-ptr nil))))))
                 (t ;; row-fetch-count > 1
                  (loop for i from 0
                        until (or (and count (= i count)))
                        do (when (null (aref prefetched-rows 0))
                             ;; reset return-code-ptr
                             (loop for return-code-ptr across column-return-code-ptrs
                                   do
                                   (dotimes (i row-fetch-count)
                                     (setf
				      #+allegro
				      (ff:fslot-value return-code-ptr i)
				      #+lispworks
				      (fli::foreign-aref return-code-ptr i)
				      +OCI-NO-DATA+)))
                             (setf result (%oci-stmt-fetch oci-stmt oci-error row-fetch-count 
                                                           *wait-function*)))
                        collect
                        (loop for data-ptr across column-data-ptrs
                              for sql-type across column-sql-types
                              for out-len-ptr across column-out-len-ptrs
                              for return-code-ptr across column-return-code-ptrs
                              for precision across column-precisions
                              for data-size across column-data-sizes
                              for j from 0 ; column count is zero based in lisp
                              collect 
                              (cond ((< 0 data-size +max-precision+)
                                     (if-let (list (aref prefetched-rows j))
                                       (pop (aref prefetched-rows j))
                                       (let ((data (read-data data-ptr sql-type
                                                              out-len-ptr return-code-ptr
                                                              data-size row-fetch-count)))
                                         (when (and (null data) (zerop i))
					   (return-from db-fetch-query-results nil))
					 (setf (aref prefetched-rows j) (cdr data))
                                         (car data))))
                                    ((zerop (%get-long out-len-ptr))
                                     nil)
                                    (t #+ignore-yet
                                       (read-data-in-chunks hstmt j data-ptr c-type sql-type
                                                            out-len-ptr nil))))
                        until (and (null (aref prefetched-rows 0))
                                   (= result +OCI-NO-DATA+))))))
         query)))))

#+lispworks
(defmacro without-interrupts (&body body)
  `(mp:without-preemption ,@body))

#+allegro
(defmacro without-interrupts (&body body)
  `(mp:without-scheduling ,@body))

#+cormanlisp
(defmacro without-interrupts (&body body)
  `(progn ,@body))

#+pcl
(defmacro without-interrupts (&body body)
  `(pcl::without-interrupts ,@body))

(defmethod db-query ((database oci-database) query-expression &optional flatp)
  (let ((free-query
         ;; make it thread safe 
         (get-free-query database)))
    (setf (sql-expression free-query) query-expression)
    (unwind-protect
      (progn
        (%db-execute free-query query-expression)
        (%initialize-query free-query)
        (db-fetch-query-results free-query nil flatp))
      (db-close-query free-query))))

(defmethod %db-execute ((database oci-database) sql-expression &key &allow-other-keys)
  (%db-execute (get-free-query database) sql-expression))

(defmethod %db-execute ((query oci-query) sql-expression &key &allow-other-keys)
  (when *trace-sql*
    (format (if (streamp *trace-sql*) *trace-sql* *trace-output*)
            "~&~A;~%" sql-expression))
  (with-slots (oci-service-context oci-environment oci-error) (sql::query-database query)
    (with-slots (oci-stmt statement-type) query
      (unless oci-stmt
        (setf oci-stmt (%new-statement-handle oci-environment oci-error))) 
      (setf (sql-expression query) sql-expression)
      (%oci-stmt-prepare oci-stmt oci-error sql-expression)
      (setf statement-type
                  (%with-temporary-allocation ((stmt-type-ptr :unsigned-short))
                    (%oci-attr-get oci-stmt +OCI-HTYPE-STMT+ stmt-type-ptr (%null-ptr)
                                   +oci-attr-stmt-type+ oci-error)
                    (%get-signed-word stmt-type-ptr)))
            (%oci-stmt-execute oci-service-context oci-stmt oci-error
                               (if (= statement-type +oci-stmt-select+) 0 1)
                               *wait-function*)
            #+ignore
            (%oci-stmt-execute oci-service-context oci-stmt oci-error 0)
      query)))

(defmethod get-free-query ((database oci-database))
  "get-free-query finds or makes a nonactive query object, and then sets it to active.
This makes the functions db-execute-command and db-query thread safe."
  (with-slots (queries) database
    (or (without-interrupts ;; not context switch allowed here 
         (let ((inactive-query (find-if (lambda (query)
                                          (not (query-active-p query)))
                                        queries)))
           ;;(describe inactive-query *local-output*)
           (when inactive-query 
             (with-slots (row-fetch-count
                          prefetched-rows
                          %no-data-p
                          column-count
                          column-names
                          column-sql-types column-data-ptrs
                          column-out-len-ptrs
                          column-return-code-ptrs
                          column-precisions column-data-sizes
                          column-scales column-nullables-p)
                         inactive-query
               (setf column-count 0
                     %no-data-p nil
                     (fill-pointer prefetched-rows) 0
                     (fill-pointer column-names) 0
                     ;;(fill-pointer column-c-types) 0
                     (fill-pointer column-sql-types) 0
                     (fill-pointer column-data-ptrs) 0
                     (fill-pointer column-out-len-ptrs) 0
                     (fill-pointer column-return-code-ptrs) 0
                     (fill-pointer column-precisions) 0
                     (fill-pointer column-data-sizes) 0
                     (fill-pointer column-scales) 0
                     (fill-pointer column-nullables-p) 0))
             (setf (query-active-p inactive-query) t))
           inactive-query))
        (let ((new-query (make-instance 'oci-query
                                        :database database
                                        :active-p t)))
          (push new-query queries)
          new-query))))

(defmethod db-execute-command ((database oci-database) sql-string)
  (db-execute-command (get-free-query database) sql-string))

(defmethod db-execute-command ((query oci-query) sql-string)
  (when *trace-sql*
    (format (if (streamp *trace-sql*) *trace-sql* *trace-output*)
            "~&~A;~%" sql-string))
  (with-slots (oci-stmt statement-type database) query
    (with-slots (oci-environment oci-error oci-service-context) database
      (unless oci-stmt (setf oci-stmt (%new-statement-handle oci-environment oci-error))) 
      (unwind-protect 
          (progn
            (%oci-stmt-prepare oci-stmt oci-error sql-string)
            (setf statement-type
                  (%with-temporary-allocation ((stmt-type-ptr :unsigned-short))
                    (%oci-attr-get oci-stmt +OCI-HTYPE-STMT+ stmt-type-ptr (%null-ptr)
                                   +oci-attr-stmt-type+ oci-error)
                    (%get-signed-word stmt-type-ptr)))
            (%oci-stmt-execute oci-service-context oci-stmt oci-error
                               (if (= statement-type +oci-stmt-select+) 0 1)
                               *wait-function*)))
        (db-close-query query))))

(defmethod %initialize-query ((database oci-database) &optional arglen col-positions)
  (%initialize-query (db-query-object database) arglen col-positions))

(defparameter *max-row-fetch-count* 1024)

(defmethod %initialize-query ((query oci-query) &optional arglen col-positions)
  (with-slots (database oci-stmt
                        row-fetch-count prefetched-rows
                        column-count column-names ;; column-c-types
                        column-sql-types column-def-ptrs
                        column-data-ptrs column-out-len-ptrs column-return-code-ptrs
                        column-precisions column-data-sizes column-scales column-nullables-p) 
      query
    (with-slots (oci-error) database
      (setf column-count 0)
      (let ((total-size 0))
        (labels ((initialize-column
                  (col-nr)
                  (multiple-value-bind (exists-p name oci-type precision
						 data-size scale nullable-p)
                      (%describe-column oci-stmt oci-error (1+ col-nr))
                    (when exists-p
                      (incf total-size (* (1+ data-size) row-fetch-count))
                      (when (= oci-type +SQLT-NUM+)
                        (setf oci-type +SQLT-VNU+))
                      ;; allocate space to bind result rows to
                      (multiple-value-bind (data-ptr def-ptr out-len-ptr return-code-ptr size #+ignore long-p)
                          (%allocate-bindings oci-type precision data-size row-fetch-count)
                        (unless nil #+not-yet long-p ;; if long-p we fetch in chunks with %sql-get-data
                          (%bind-column oci-stmt def-ptr oci-error col-nr oci-type
                                        data-ptr (1+ size) out-len-ptr return-code-ptr))
                        (vector-push-extend () prefetched-rows)
                        (vector-push-extend name column-names) 
                        (vector-push-extend oci-type column-sql-types)
                        (vector-push-extend precision column-precisions)
                        (vector-push-extend def-ptr column-def-ptrs)
                        (vector-push-extend data-size column-data-sizes)
                        (vector-push-extend scale column-scales)
                        (vector-push-extend nullable-p column-nullables-p)
                        (vector-push-extend data-ptr column-data-ptrs)
                        (vector-push-extend out-len-ptr column-out-len-ptrs)
                        (vector-push-extend return-code-ptr column-return-code-ptrs))))))
	  (let ((max-size 0))
	    (cond (col-positions
		   (dolist (col-nr col-positions)
		     (setf max-size
			   (max max-size
				(%get-column-data-size oci-stmt oci-error (1+ col-nr)))))
		   (setf row-fetch-count (min (floor 32767 (1+ max-size)) *max-row-fetch-count*))
		   (dolist (col-nr col-positions)
		     (initialize-column col-nr)))
		  (arglen
		   (loop for col-nr from 0 to (1- arglen)
			 for size = (%get-column-data-size oci-stmt oci-error (1+ col-nr))
			 while size
			 do (setf max-size (max max-size size)))
		   (setf row-fetch-count (min (floor 32767 (1+ max-size)) *max-row-fetch-count*))
		   (loop for col-nr from 0 to (1- arglen)
			 while (initialize-column col-nr)
			 do (incf column-count)))
		  (t
		   (loop for col-nr from 0
			 for size = (%get-column-data-size oci-stmt oci-error (1+ col-nr))
			 while size
			 do (setf max-size (max max-size size)))
		   (setf row-fetch-count (min (floor 32767 (1+ max-size)) *max-row-fetch-count*))
		   (loop for col-nr from 0
			 while (initialize-column col-nr)
			 do (incf column-count))
		   #+ignore
		   (dotimes (col-nr column-count)
		     ;; get column information
		     (initialize-column col-nr))))))))
    query))

(defmethod db-close-query ((query oci-query) &key drop-p)
  (with-slots (oci-stmt statement-type column-data-ptrs column-out-len-ptrs column-return-code-ptrs
                        column-def-ptrs database) query
    (let ((count (fill-pointer column-data-ptrs)))
      (when (not (zerop count))
        (dotimes (col-nr count)
          (let ((data-ptr (aref column-data-ptrs col-nr))
                (out-len-ptr (aref column-out-len-ptrs col-nr))
                (def-ptr (aref column-def-ptrs col-nr))
                (return-code-ptr (aref column-return-code-ptrs col-nr)))
            (when data-ptr (%dispose-ptr data-ptr)) ; we *did* allocate them
            (when out-len-ptr (%dispose-ptr out-len-ptr))
            (when def-ptr (%dispose-ptr def-ptr))
            (when return-code-ptr (%dispose-ptr return-code-ptr)))))
      (cond ((null oci-stmt)
             nil)
            (t
             (%oci-stmt-fetch oci-stmt (oci-error database) 0 nil)))
      (setf (query-active-p query) nil)))
  query)

(defmethod %read-query-data ((database oci-database) ignore-columns)
  (%read-query-data (db-query-object database) ignore-columns))

;; does not work yet for row-fetch-count = 1!
(defmethod %read-query-data ((query oci-query) ignore-columns)
  (with-slots (database oci-stmt prefetched-rows row-fetch-count  
                        column-count column-c-types column-sql-types
                        column-data-ptrs column-out-len-ptrs column-data-sizes
                        column-return-code-ptrs column-precisions
                        %no-data-p)
              query
    (when (null (aref prefetched-rows 0))
      (when %no-data-p (return-from %read-query-data nil))
      ;; reset return-code-ptr
      (loop for return-code-ptr across column-return-code-ptrs
            do (dotimes (i row-fetch-count)
                 (setf #+allegro
		       (ff:fslot-value return-code-ptr i)
		       #+ignore
		       (sys:memref-int return-code-ptr 0 i :unsigned-short)
		       #+lispworks
		       (fli::foreign-aref return-code-ptr i)
		       +OCI-NO-DATA+)))
      (setf %no-data-p  
            (= +OCI-NO-DATA+
	       (%oci-stmt-fetch oci-stmt (oci-error database) row-fetch-count *wait-function*))))
    (values
     (loop for col-nr from 0 to (- column-count (if (eq ignore-columns :last) 2 1))
           collect
           (let ((precision (aref column-precisions col-nr)))
             (cond (t (if-let (list (aref prefetched-rows col-nr))
                        (pop (aref prefetched-rows col-nr))
                        (let ((data (read-data (aref column-data-ptrs col-nr)
                                               (aref column-sql-types col-nr)
                                               (aref column-out-len-ptrs col-nr)
                                               (aref column-return-code-ptrs col-nr)
                                               (aref column-data-sizes col-nr)
                                               row-fetch-count)))
                          (when (null data) (return-from %read-query-data nil))
                          (setf (aref prefetched-rows col-nr) (cdr data))
                          (car data))))
                   #+not-yet
                   (t
                    (read-data-in-chunks hstmt col-nr
                                         (aref column-data-ptrs col-nr) 
                                         (aref column-c-types col-nr)
                                         (aref column-sql-types col-nr)
                                         (aref column-out-len-ptrs col-nr)
                                         nil)))))
     t)))

(defmethod db-map-query ((database oci-database) type function query-exp)
  (db-map-query (get-free-query database) type function query-exp))

(defmethod db-map-query ((query oci-query) type function query-exp)
  (declare (ignore type)) ; preliminary. Do a type coersion here
  (%db-execute query (sql::sql-string query-exp))
  (unwind-protect
    (progn
      (%initialize-query query)
      ;; the main loop
      (loop for data = (%read-query-data query nil)
            while data
            do (apply function data)))
    ;; dispose of memory and set query inactive or get rid of it
    (db-close-query query)))

(defmethod db-map-bind-query ((query oci-query) type function &rest parameters)
  (declare (ignore type)) ; preliminary. Do a type coersion here
  (unwind-protect
    (progn
      (apply #'%db-bind-execute query parameters)
      ;; the main loop
      (loop with data
            while (setf data (%read-query-data query nil))
            do (apply function data)))
    ;; dispose of memory and set query inactive or get rid of it
    (%db-reset-query query)))

;; prepared queries

; EOF
